#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "gauss.h"

crossing *crossings;
int cyclic;
int reversible;

void addcode(code *, int *, codeelem *, int *, codeelem *, int, int, int);
code firstrot(code);
void fixnums(code, int);

void initcrossings(int sz) { // is called by gencodes/genallcodes
  static int crosssz = 0;
  if (crosssz >= sz)
    return;
  if (crosssz > 0)
    free (crossings);
  crossings = malloc(6 * sz * sizeof(crossing));
  int i, j;
  for (i = 0; i < sz; i++) {
    for (j = 0; j < 6; j++) {
      crossings[6 * i + j].n = i;
      crossings[6 * i + j].p = j/3;
      crossings[6 * i + j].s = j % 3 - 1;
    }
  }
  crosssz = sz;
}

code *genallcodes(int sz, knot *n) {
  // generates all signed codes with n<=sz
  initcrossings(sz);
  code *codes[sz + 1], *allcodes;
  int nums[sz + 1], i, m = 0;
  for (i = 0; i <= sz; i++) {
    codes[i] = gencodes(i, nums + i);
    m += nums[i];
  }
  allcodes = malloc(m * sizeof(code));
  *n = ((knot) m) << sz;
  
  m = 0;
  for (i = 0; i <= sz; i++) {
    memcpy(allcodes + m, codes[i], nums[i] * sizeof(code));
    m += nums[i];
    free(codes[i]);
  }
  return allcodes;
}

code getcode(code *list, int sz, knot n) { // generates oriented code
  code c = list[(unsigned int) (n >> sz)];
  int o = (int) (n & ((1 << sz) - 1));
  if (o >= 1 << c.n) {
    c.n = -1;
    return c;
  }
  code r;
  r.n = c.n;
  r.code = malloc(2 * r.n * sizeof(code));
  int l;
  for (l = 0; l < 2 * r.n; l++)
    r.code[l] = c.code[l] +
      (((o & (1 << (r.n - 1 - c.code[l] / 6))) > 0) ? 1 : -1);
  return r;
}

int getsigns(code c) { // get signs of the crossings of c
  int r = 0, i;
  for (i = 0; i < 2 * c.n; i++)
    if (c.code[i] % 3 == 2)
      r |= 1 << (c.n - 1 - c.code[i] / 6);
  return r;
}

code *gencodes(int sz, int *n) {
  initcrossings(sz);
  code *codes = malloc((1 << 24) * sizeof(code));
  codeelem *elems = malloc(2 * sz * sizeof(codeelem));
  int i;

  for (i = 0; i < 2 * sz; i++) // create array of elements to choose from
    elems[i] = 3 * i + 1;  // only the ones with unsigned crossings

  *n = 0;
  int *used = calloc(2 * sz, sizeof(int));
  codeelem *c = malloc(2 * sz * sizeof(codeelem));
  addcode(codes, n, elems, used, c, sz, 2, 0);

  free(elems);
  free(c);
  free(used);
  codes = realloc(codes, *n * sizeof(code));

  return codes;
}

void addcode(code *codes, int *cn, codeelem *elems, int *used, codeelem *c,
	     int sz, int max, int n) { // recursive; does work for gencodes
  if (*cn >= 1 << 24) {
    fprintf(stderr, "Array space exhausted\n");
    return;
  }
  int i;
  if (n == sz * 2) { // code is complete and ready to add
    for (i = 1; i < 2 * sz; i++) { // eliminate codes with R1
      if (c[i] / 6 == c[i - 1] / 6)
	return;
    }
    if (cyclic && sz > 0 && c[0] / 6 == c[2 * sz - 1] / 6)
      return;
    code cyc, rot;
    cyc.code = c;
    cyc.n = sz;
    rot = first(cyc);
    if (rot.code != cyc.code) { // a cyclically equivalent code will have
      free(rot.code);           // been already made
      return;
    }
    codes[*cn].code = malloc(2 * sz * sizeof(codeelem));
    memcpy(codes[*cn].code, c, 2 * sz * sizeof(codeelem));
    codes[(*cn)++].n = sz;
    return;
  }

  for (i = 0; i < max; i++) { // try different choices for next element
    if (!used[i]) {
      used[i] = 1;
      c[n] = elems[i];
      addcode(codes, cn, elems, used, c, sz,
	      (i >= max - 2 && max < 2 * sz) ? max + 2 : max, n + 1);
      used[i] = 0;
    }
  }
}

int compare(code a, code b) { return compare2(a, b, 1); }

int compare2(code a, code b, int s) { // -1 if a < b, 1 if a > b, 0 if a == b
  if (a.n < b.n)                      // if s is 0, ignore signs of crossings
    return -1;
  if (a.n > b.n)
    return 1;
  int i;
  for (i = 0; i < a.n * 2; i++) {
    if (a.code[i] / 3 < b.code[i] / 3)
      return -1;
    if (a.code[i] / 3 > b.code[i] / 3)
      return 1;
  }
  if (s) {
    for (i = 0; i < a.n * 2; i++) {
      if (a.code[i] % 3 < b.code[i] % 3)
	return -1;
      if (a.code[i] % 3 > b.code[i] % 3)
	return 1;
    }
  }
  return 0;
}

int find(code c, code *list, int n, int m) { // find index of c in list
  if (m <= n)
    return -1;
  int cmp = compare2(c, list[(m + n)/2], 0);  // ignores orientations
  if (cmp == 0)
    return (m + n)/2;
  if (cmp < 0)
    return find(c, list, n, (m + n)/2);
  return find(c, list, (n + m + 1)/2, m);
}

code rotate(code c) { // rotate elements of c by one to the left
  code r;             // and relabel them from 0
  r.n = c.n;
  r.code = malloc(2 * r.n * sizeof(codeelem));
  int i, renames[r.n], n = 0;
  for (i = 0; i < r.n; i++)
    renames[i] = -1;
  for (i = 1; i < 2 * r.n; i++) {
    int op, m;
    op = c.code[i] % 6;
    m = c.code[i] / 6;
    if (renames[m] != -1)
      r.code[i - 1] = 6 * renames[m] + op;
    else {
      r.code[i - 1] = 6 * n + op;
      renames[m] = n++;
    }
  }
  r.code[2 * r.n - 1] = c.code[0] % 6 + 6 * renames[c.code[0] / 6];
  return r;
}

code first(code a) { // returns the rotated/reversed form of a that we'll use
  if (!reversible)
    if (cyclic)
      return firstrot(a);
    else
      return a;
  code b;
  b.n = a.n;
  b.code = malloc(2 * b.n * sizeof(codeelem));
  int i;
  for (i = 0; i < 2 * a.n; i++)
    b.code[i] = a.code[2 * a.n - i - 1];
  fixnums(b, b.n);
  if (cyclic) {
    code ar = firstrot(a);
    code br = firstrot(b);
    if (b.code != br.code)
      free(b.code);
    if (compare(br, ar) < 0) {
      if (ar.code != a.code)
	free(ar.code);
      return br;
    }
    free(br.code);
    return ar;
  }
  if (compare(b, a) < 0)
    return b;
  free(b.code);
  return a;
}
    

code firstrot(code a) { // return the first (ordered by compare()) of the
  code c = a;           // rotations of a
  code first = c;
  code prev = c;
  int i;

  for (i = 1; i < 2 * c.n; i++) {
    prev = c;
    c = rotate(prev);
    if (i != 1 && prev.code != first.code)
      free(prev.code);
    if (compare(c, first) < 0) {
      if (first.code != a.code)
	free(first.code);
      first = c;
    }
  }
  if (c.code != first.code && c.code != a.code)
    free(c.code);

  if (compare(first, a) == 0 && first.code != a.code) { // works nicer this way
    free(first.code);
    return a;
  }
  return first;
}

void printcode(code c) {
  int i;
  for (i = 0; i < 2 * c.n; i++) {
    printf((crossings[c.code[i]].p == U) ? "U" : "O");
    printf("%d", c.code[i] / 6 + 1);
    if (crossings[c.code[i]].s != 0)
      printf((crossings[c.code[i]].s > 0) ? "+" : "-");
  }
  printf("\n");
}

char *sprintcode(code c) {
  int i, j = 0;
  char *s = malloc((8 * c.n + 1) * sizeof(char)); // assumes < 100 crossings
  s[0] = '\0';
  for (i = 0; i < 2 * c.n; i++) {
    j += sprintf(s + j, (crossings[c.code[i]].p == U) ? "U" : "O");
    j += sprintf(s + j, "%d", c.code[i] / 6 + 1);
    if (crossings[c.code[i]].s != 0)
      j += sprintf(s + j, (crossings[c.code[i]].s > 0) ? "+" : "-");
  }
  return s;
}

code parsecode(char *s) {
  int i, j, l;
  code c;
  l = strlen(s);
  c.code = malloc(l * sizeof(codeelem));
  for (i = 0, j = -1; i < l; i++)
    switch(s[i]) {
    case 'o':
    case 'O':
      c.code[++j] = 3 * O;
      break;
    case 'u':
    case 'U':
      c.code[++j] = 3 * U;
      break;
    case '-':
      break;
    case '+':
      c.code[j] += 2;
      break;
    default:
      if isdigit(s[i]) {
	int k;
	for (k = i; isdigit(s[k]); k++);
	//char tmp = s[k];
	//s[k] = '\0';
	c.code[j] += 6 * (atoi(s + i) - 1);
	//s[k] = tmp;
	i = k - 1;
      }
    }
  c.n = (j + 1) / 2;
  c.code = realloc(c.code, 2 * c.n * sizeof(codeelem));
  return c;
}

void fixnums(code c, int nn) { // fixes numbering within c
  int renames[nn], i, n = 0;   // nn is number of crossings
  for (i = 0; i < nn; i++)
    renames[i] = -1;
  for (i = 0; i < 2 * c.n; i++) {
    int m = c.code[i] / 6;
    int op = c.code[i] % 6;
    if (renames[m] != -1)
      c.code[i] = 6 * renames[m] + op;
    else {
      c.code[i] = 6 * n + op;
      renames[m] = n++;
    }
  }
}

code simplify(code c) {
  int i;
  code b, r;
  r.n = c.n;
  r.code = malloc(2 * c.n * sizeof(codeelem));
  b.code = malloc(2 * c.n * sizeof(codeelem));
  memcpy(r.code, c.code, 2 * c.n * sizeof(codeelem));
  do {        // repeat until no more doublets are left
    b.n = r.n;
    memcpy(b.code, r.code, 2 * r.n * sizeof(codeelem));
    r.n = 0;
    for (i = 0; i < 2 * b.n; i++) {
      if (cyclic && (i == 0 || i == 2 * b.n - 1)
	  && b.code[0] / 6 == b.code[2 * b.n - 1] / 6)
	continue;
      if (i != 0 && b.code[i] / 6 == b.code[i - 1] / 6)
	continue;
      if (i != 2 * b.n - 1 && b.code[i] / 6 == b.code[i + 1] / 6)
	continue;
      r.code[r.n++] = b.code[i];
    }
    r.n /= 2;
  } while (r.n != b.n);
  free(b.code);

  fixnums(r, c.n + 2); // +2 for R2 case
  code f = first(r);
  if (f.code != r.code)
    free(r.code);
  if (compare(f, c) == 0) {
    free(f.code);
    return c;
  }
  return f;
}
